/* z80.h */

/*
 * (C) Copyright 1989-1999
 * All Rights Reserved
 *
 * Alan R. Baldwin
 * 721 Berkeley St.
 * Kent, Ohio  44240
 */

/*)BUILD
	$(PROGRAM) =	ASZ80
	$(INCLUDE) = {
		ASXXXX.H
		Z80.H
	}
	$(FILES) = {
		Z80EXT.C
		Z80MCH.C
		Z80ADR.C
		Z80PST.C
		ASMAIN.C
		ASLEX.C
		ASSYM.C
		ASSUBR.C
		ASEXPR.C
		ASDATA.C
		ASLIST.C
		ASOUT.C
	}
	$(STACK) = 3000
*/

/*
 * Indirect Addressing delimeters
 */
#define	LFIND	'('
#define RTIND	')'

/*
 * Registers
 */
#define B	0
#define C	1
#define D	2
#define E	3
#define H	4
#define L	5
#define A	7

#define F	6
#define I	0x47
#define R	0x4f

#define BC	0
#define DE	1
#define HL	2
#define SP	3
#define AF	4
#define IX	5
#define IY	6

/* Z80 Undocumented */
#define IXH	H
#define IXL	L
#define IYH	H
#define IYL	L

/*
 * Special Addressing
 */

/* Z80 Undocumented */
#define LDOP_RLC	0
#define LDOP_RRC	1
#define LDOP_RL	2
#define LDOP_RR	3
#define LDOP_SLA	4
#define LDOP_SRA	5
#define LDOP_SLL	6
#define LDOP_SRL	7
#define LDOP_RES	0x10
#define LDOP_SET	0x18

/* DMG-SYSTEM(GAMEBOY) */
#define HLD	IX
#define HLI	IY

/*
 * Conditional definitions
 */
#define	NZ	0
#define	Z	1
#define	NC	2
#define	CS	3
#define	PO	4
#define	PE	5
#define	P	6
#define	M	7

/*
 * Symbol types
 */
enum typesym {
	S_IMMED = 30,
	S_R8,
	S_R8X,
	S_R16,
	S_R16X,
	S_CND,
	S_FLAG,
/* Z80 Undocumented */
	S_R8IX,
	S_R8IY,

/*
 * Indexing modes
 */
	S_INDB,
	S_IDC,
	S_IDBC,
#define S_INDR S_IDBC
	S_IDDE,
	S_IDHL,
	S_IDSP,
	S_INDM,
	S_IDIX,
	S_IDIY,
/* DMG-SYSTEM(GAMEBOY) */
#define S_IDHLD S_IDIX
#define S_IDHLI S_IDIY

/*
 * Instruction types
 */
	X_Z80,
	S_LD,
	S_CALL,
	S_JP,
	S_JR,
	S_RET,
	S_BIT,
	S_INC,
	S_DEC,
	S_ADD,
	S_ADC,
	S_AND,
	S_EX,
	S_PUSH,
	S_IN,
	S_OUT,
	S_RL,
	S_RST,
	S_IM,
	S_INH1,
	S_INH2,
	S_LDD,
	S_LDI,
	S_DJNZ,
	S_SUB,
	S_SBC,

/*
 * Z80 Undocumented Instructions
 */
	X_SLL,

/*
 * HD64180 Instructions
 */
	X_HD64,
	X_INH2,
	X_IN,
	X_OUT,
	X_MLT,
	X_TST,
	X_TSTIO,
	X_HD64END,

/*
 * DMG-SYSTEM(GAMEBOY) Instructions
 */
	X_DMG,
	X_LDA,
	X_LDH,
	X_LDHL,
	X_STOP,
	X_DMGEND,

/*
 * R800 Instructions
 */
	X_R800,
	X_MULU,
	X_R800END
};

struct adsym
{
	char	a_str[4];	/* addressing string */
	int	a_val;		/* addressing mode value */
};

extern enum target_cpu {
	TARGET_Z80,
	TARGET_R800,
	TARGET_HD64,
	TARGET_DMG
	/*,TARGET_Z380*/
} target;

extern	struct	adsym	R8[];
extern	struct	adsym	R8X[];
extern	struct	adsym	R16[];
extern	struct	adsym	R16X[];
extern	struct	adsym	CND[];
extern	struct	adsym	R16I[];
/* Z80 Undocumented */
extern	struct	adsym	R8IX[];
extern	struct	adsym	R8IY[];
extern	struct	adsym	LDOP[];
/* DMG-SYSTEM(GAMEBOY) */
extern	struct	adsym	R16GB[];

	/* machine dependent functions */

	/* z80adr.c */
extern	int		addr();
extern	int		admode();
extern	int		any();
extern	int		srch();

	/* z80mch.c */
extern	int		comma();
extern	int		genop();
extern	int		gixiy();
extern	VOID		machine();
extern	int		mchpcr();
extern	VOID		minit();
