/* z80adr.c */

/*
 * (C) Copyright 1989-1999
 * All Rights Reserved
 *
 * Alan R. Baldwin
 * 721 Berkeley St.
 * Kent, Ohio  44240
 */

#include <stdio.h>
#include <setjmp.h>
#include "asxxxx.h"
#include "z80.h"

/*
 * Read an address specifier. Pack the
 * address information into the supplied
 * `expr' structure. Return the mode of
 * the address.
 *
 * This addr(esp) routine performs the following addressing decoding:
 *
 *	address		mode		flag		addr		base
 *	#n		S_IMMED		0		n		NULL
 *	label		s_type		----		s_addr		s_area
 *	[REG]		S_IND+icode	0		0		NULL
 *	[label]		S_INDM		----		s_addr		s_area
 *	offset[REG]	S_IND+icode	----		offset		----
 */
int
addr(esp)
register struct expr *esp;
{
	register int c, mode, indx;

	if ((c = getnb()) == '#') {
		expr(esp, 0);
		esp->e_mode = S_IMMED;
	} else
	if (c == LFIND) {
		if ((indx = admode(R8)) != 0) {
			mode = S_INDB;
		} else
		if (target != TARGET_DMG && (indx = admode(R8IX)) != 0) {
			mode = S_INDB;
			aerr();
		} else
		if (target != TARGET_DMG && (indx = admode(R8IY)) != 0) {
			mode = S_INDB;
			aerr();
		} else
		if (target == TARGET_DMG && (indx = admode(R16GB)) != 0) {
			mode = S_INDR;
		} else
		if ((indx = admode(R16)) != 0) {
			mode = S_INDR;
		} else	
		if (target != TARGET_DMG && (indx = admode(R16I)) != 0) {
			mode = S_INDR;
		} else	
		if (target != TARGET_DMG && (indx = admode(R8X)) != 0) {
			mode = S_R8X;
			aerr();
		} else
		if ((indx = admode(R16X)) != 0) {
			mode = S_R16X;
			aerr();
		} else {
			mode = S_INDM;
			expr(esp, 0);
			esp->e_mode = mode;
		}
		if (indx) {
			esp->e_mode = mode + (indx&0xFF);
			esp->e_base.e_ap = NULL;
		}
		if ((c = getnb()) != RTIND)
			qerr();
	} else {
		unget(c);
		if ((indx = admode(R8)) != 0) {
			mode = S_R8;
		} else
		if (target != TARGET_DMG && (indx = admode(R8IX)) != 0) {
			mode = S_R8IX;
		} else
		if (target != TARGET_DMG && (indx = admode(R8IY)) != 0) {
			mode = S_R8IY;
		} else
		if (target == TARGET_DMG && (indx = admode(R16GB)) != 0) {
			mode = S_R16;
		} else
		if ((indx = admode(R16)) != 0) {
			mode = S_R16;
		} else	
		if (target != TARGET_DMG && (indx = admode(R16I)) != 0) {
			mode = S_R16;
		} else	
		if (target != TARGET_DMG && (indx = admode(R8X)) != 0) {
			mode = S_R8X;
		} else
		if ((indx = admode(R16X)) != 0) {
			mode = S_R16X;
		} else {
			mode = S_USER;
			expr(esp, 0);
			esp->e_mode = mode;
		}
		if (indx) {
			esp->e_addr = indx&0xFF;
			esp->e_mode = mode;
			esp->e_base.e_ap = NULL;
		}
		if ((c = getnb()) == LFIND) {
if (target == TARGET_DMG) {
			if ((indx=admode(R16))!=0) {
				esp->e_mode = S_INDR + (indx&0xFF);
			} else {
				aerr();
			}
} else {
			if ((indx=admode(R16I))!=0) {
				esp->e_mode = S_INDR + (indx&0xFF);
			} else {
				aerr();
			}
}
			if ((c = getnb()) != RTIND)
				qerr();
		} else {
			unget(c);
		}
	}
	return (esp->e_mode);
}

/*
 * Enter admode() to search a specific addressing mode table
 * for a match. Return the addressing value on a match or
 * zero for no match.
 */
int
admode(sp)
register struct adsym *sp;
{
	register char *ptr;
	register int i;
	register char *ips;

	ips = ip;
	unget(getnb());

	i = 0;
	while ( *(ptr = &sp[i].a_str[0]) ) {
		if (srch(ptr)) {
			return(sp[i].a_val);
		}
		i++;
	}
	ip = ips;
	return(0);
}

/*
 *      srch --- does string match ?
 */
int
srch(str)
register char *str;
{
	register char *ptr;
	ptr = ip;

#ifdef	CASE_SENSITIVE
	while (*ptr && *str) {
		if (*ptr != *str)
			break;
		ptr++;
		str++;
	}
	if (*ptr == *str) {
		ip = ptr;
		return(1);
	}
#else
	while (*ptr && *str) {
		if (ccase[*ptr & 0x007F] != ccase[*str & 0x007F])
			break;
		ptr++;
		str++;
	}
	if (ccase[*ptr & 0x007F] == ccase[*str & 0x007F]) {
		ip = ptr;
		return(1);
	}
#endif

	if (!*str)
		if (any(*ptr," \t\n,);")) {
			ip = ptr;
			return(1);
		}
	return(0);
}

/*
 *      any --- does str contain c?
 */
int
any(c,str)
char    c, *str;
{
	while (*str)
		if(*str++ == c)
			return(1);
	return(0);
}

/*
 * Registers
 */

struct	adsym	R8[] = {
    {	"b",	B|0x400	},
    {	"c",	C|0x400	},
    {	"d",	D|0x400	},
    {	"e",	E|0x400	},
    {	"h",	H|0x400	},
    {	"l",	L|0x400	},
    {	"a",	A|0x400	},
    {	"",	0000	}
};

struct	adsym	R8X[] = {
    {	"i",	I|0x400	},
    {	"r",	R|0x400	},
    {	"f",	F|0x400	},
    {	"",	0000	}
};

struct	adsym	R16[] = {
    {	"bc",	BC|0x400	},
    {	"de",	DE|0x400	},
    {	"hl",	HL|0x400	},
    {	"sp",	SP|0x400	},
    {	"",	0000	}
};

struct	adsym	R16I[] = {
    {	"ix",	IX|0x400	},
    {	"iy",	IY|0x400	},
    {	"",	0000	}
};

struct	adsym	R16X[] = {
    {	"af",	AF|0x400	},
    {	"af'",	AF|0x400	},
    {	"",	0000	}
};

/* Z80 Undocumented */

struct	adsym	R8IX[] = {
    {	"ixh",	IXH|0x400	},
    {	"ixl",	IXL|0x400	},
    {	"",	0000	}
};

struct	adsym	R8IY[] = {
    {	"iyh",	IYH|0x400	},
    {	"iyl",	IYL|0x400	},
    {	"",	0000	}
};

struct	adsym	LDOP[] = {
    {	"rlc",	LDOP_RLC|0x400	},
    {	"rrc",	LDOP_RRC|0x400	},
    {	"rl",	LDOP_RL|0x400	},
    {	"rr",	LDOP_RR|0x400	},
    {	"sla",	LDOP_SLA|0x400	},
    {	"sra",	LDOP_SRA|0x400	},
    {	"sll",	LDOP_SLL|0x400	},
    {	"srl",	LDOP_SRL|0x400	},
    {	"res",	LDOP_RES|0x400	},
    {	"set",	LDOP_SET|0x400	},
    {	"",	0000	}
};

/* DMG-SYSTEM(GAMEBOY) */
struct	adsym	R16GB[] = {
    {	"hl-",	HLD|0x400},
    {	"hl+",	HLI|0x400},
    {	"hld",	HLD|0x400},
    {	"hli",	HLI|0x400},
    {	"",	0000	}
};


/*
 * Conditional definitions
 */

struct	adsym	CND[] = {
    {	"NZ",	NZ|0x400	},
    {	"Z",	Z |0x400	},
    {	"NC",	NC|0x400	},
    {	"C",	CS|0x400	},
    {	"PO",	PO|0x400	},
    {	"PE",	PE|0x400	},
    {	"P",	P |0x400	},
    {	"M",	M |0x400	},
    {	"",	0000	}
};
