/* z80mch.c */

/*
 * (C) Copyright 1989-1999
 * All Rights Reserved
 *
 * Alan R. Baldwin
 * 721 Berkeley St.
 * Kent, Ohio  44240
 */

#include <stdio.h>
#include <setjmp.h>
#include "asxxxx.h"
#include "z80.h"

char	imtab[3] = { 0x46, 0x56, 0x5E };
enum	target_cpu target;

/*
 * Process a machine op.
 */
VOID
machine(mp)
struct mne *mp;
{
	register int op, t1, t2;
	struct expr e1, e2;
	int rf, v1, v2;

	clrexpr(&e1);
	clrexpr(&e2);
	op = mp->m_valu;
	rf = mp->m_type;
	switch (target) {
		case TARGET_Z80:
			/* DMG-SYSTEM(GAMEBOY) Instructions */
			if ((rf > X_DMG) && (rf < X_DMGEND)) rf = 0;
			/* R800 Instructions */
			else if ((rf > X_R800) && (rf < X_R800END)) rf = 0;
			/* HD64180 Instructions */
			else if ((rf > X_HD64) && (rf < X_HD64END)) rf = 0;
			break;
		case TARGET_R800:
			/* DMG-SYSTEM(GAMEBOY) Instructions */
			if ((rf > X_DMG) && (rf < X_DMGEND)) rf = 0;
			/* HD64180 Instructions */
			else if ((rf > X_HD64) && (rf < X_HD64END)) rf = 0;
			/* Z80 Instructions */
			else if (rf == X_SLL) rf = 0;
			break;
		case TARGET_HD64:
			/* DMG-SYSTEM(GAMEBOY) Instructions */
			if ((rf > X_DMG) && (rf < X_DMGEND)) rf = 0;
			/* R800 Instructions */
			else if ((rf > X_R800) && (rf < X_R800END)) rf = 0;
			break;
		case TARGET_DMG:
			if (rf == S_INH2) {
				if (op == 0x4D) {
					/* remap 'RETI' */
					rf = S_INH1;
					op = 0xD9;
				} else
					rf = 0;
			}
			else if (rf == S_EX) rf = 0;
			else if (rf == S_IN) rf = 0;
			else if (rf == S_OUT) rf = 0;
			else if (rf == S_DJNZ) rf = 0;
			/* R800 Instructions */
			else if ((rf > X_R800) && (rf < X_R800END)) rf = 0;
			/* HD64180 Instructions */
			else if ((rf > X_HD64) && (rf < X_HD64END)) rf = 0;
			/* Z80 Instructions */
			else if (rf == X_SLL) rf = 0;
			break;
	}
	switch (rf) {

	case S_INH1:
		outab(op);
		break;

	case S_INH2:
		outab(0xED);
		outab(op);
		break;

	case S_LDI:
		if (target != TARGET_DMG) {
			outab(0xED);
			outab(op);
			break;
		} else {
			/*
			 * 0x22 : LDI (HL),A
			 * 0x2A : LDI A,(HL)
			 */
			t1 = addr(&e1);
			comma();
			t2 = addr(&e2);
			v1 = e1.e_addr;
			v2 = e2.e_addr;
			if ((t1 == S_R8) && (v1 == A) && (t2 == S_IDHL)) {
				outab(0x2A);
				break;
			}
			if ((t2 == S_R8) && (v2 == A) && (t1 == S_IDHL)) {
				outab(0x22);
				break;
			}
			aerr();
			break;
		}

	case S_LDD:
		if (target != TARGET_DMG) {
			outab(0xED);
			outab(op);
			break;
		} else {
			/*
			 * 0x32 : LDD (HL-),A
			 * 0x3A : LDD A,(HL-)
			 */
			t1 = addr(&e1);
			comma();
			t2 = addr(&e2);
			v1 = e1.e_addr;
			v2 = e2.e_addr;
			if ((t1 == S_R8) && (v1 == A) && (t2 == S_IDHL)) {
				outab(0x3A);
				break;
			}
			if ((t2 == S_R8) && (v2 == A) && (t1 == S_IDHL)) {
				outab(0x32);
				break;
			}
			aerr();
			break;
		}

	case S_RET:
		if (more()) {
			if ((v1 = admode(CND)) != 0) {
				outab(op | v1<<3);
			} else {
				qerr();
			}
		} else {
			outab(0xC9);
		}
		break;

	case S_PUSH:
		t1 = addr(&e1);
		v1 = e1.e_addr;
		if ((t1 == S_R16X) && (v1 == AF)) {
			outab(op+0x30);
			break;
		} else
		if ((t1 == S_R16) && (v1 != SP)) {
			if (v1 != gixiy(v1)) {
				outab(op+0x20);
				break;
			}
			outab(op | v1<<4);
			break;
		}
		aerr();
		break;

	case S_RST:
		v1 = absexpr();
		if (v1 & ~0x38) {
			aerr();
			v1 = 0;
		}
		outab(op|v1);
		break;

	case S_IM:
		expr(&e1, 0);
		abscheck(&e1);
		if (e1.e_addr > 2) {
			aerr();
			e1.e_addr = 0;
		}
		outab(op);
		outab(imtab[e1.e_addr]);
		break;

	case S_BIT:
		expr(&e1, 0);
		t1 = 0;
		v1 = e1.e_addr;
		if (v1 > 7) {
			++t1;
			v1 &= 0x07;
		}
		op |= (v1<<3);
		comma();
		addr(&e2);
		abscheck(&e1);
		if (genop(0xCB, op, &e2, 0) || t1)
			aerr();
		break;

	case X_SLL:
	case S_RL:
		t1 = 0;
		t2 = addr(&e2);
		if (more()) {
			if ((t2 != S_R8) || (e2.e_addr != A))
				++t1;
			comma();
			t2 = addr(&e2);
		}
		if (genop(0xCB, op, &e2, 0) || t1)
			aerr();
		break;

	case S_AND:
	case S_SUB:
		t1 = 0;
		t2 = addr(&e2);
		if (more()) {
			if ((t2 != S_R8) || (e2.e_addr != A))
				++t1;
			comma();
			t2 = addr(&e2);
		}
		if (genop(0, op, &e2, 1) || t1)
			aerr();
		break;

	case S_ADD:
	case S_ADC:
	case S_SBC:
		t1 = addr(&e1);
		t2 = 0;
		if (more()) {
			comma();
			t2 = addr(&e2);
		}
		if (t2 == 0) {
			if (genop(0, op, &e1, 1))
				aerr();
			break;
		}
		if ((t1 == S_R8) && (e1.e_addr == A)) {
			if (genop(0, op, &e2, 1))
				aerr();
			break;
		}
if (target != TARGET_DMG) {
		if ((t1 == S_R16) && (t2 == S_R16)) {
			if (rf == S_ADD)
				op = 0x09;
			if (rf == S_ADC)
				op = 0x4A;
			if (rf == S_SBC)
				op = 0x42;
			v1 = e1.e_addr;
			v2 = e2.e_addr;
			if ((v1 == HL) && (v2 <= SP)) {
				if (rf != S_ADD)
					outab(0xED);
				outab(op | (v2<<4));
				break;
			}
			if (rf != S_ADD) {
				aerr();
				break;
			}
			if ((v1 == IX) && (v2 != HL) && (v2 != IY)) {
				if (v2 == IX) v2 = HL;
				outab(0xDD);
				outab(op | (v2<<4));
				break;
			}
			if ((v1 == IY) && (v2 != HL) && (v2 != IX)) {
				if (v2 == IY) v2 = HL;
				outab(0xFD);
				outab(op | (v2<<4));
				break;
			}
		}
} else {
		if ((t1 == S_R16) && (t2 == S_R16)) {
			v1 = e1.e_addr;
			v2 = e2.e_addr;
			if ((v1 == HL) && (v2 <= SP) && (rf == S_ADD)) {
				outab(0x09 | (v2<<4));
				break;
			}
		}
		/*
		 * 0xE8 : ADD SP,#n
		 */
		if ((rf == S_ADD) && (t1 == S_R16) && (e1.e_addr == SP) && (t2 == S_IMMED)) {
			outab(0xE8);
			outrb(&e2,0);
			break;
		}
}
		aerr();
		break;

	case S_LD:
		t1 = addr(&e1);
		comma();
		if ((target == TARGET_Z80) || (target == TARGET_HD64)) {
			if ((t1 == S_R8) && ((v1 = admode(LDOP)) != 0)) {
				if (((v1 & 0xFF) == LDOP_RES) || ((v1 & 0xFF) == LDOP_SET)) {
					expr(&e2, 0);
					t2 = 0;
					v2 = e2.e_addr;
					op = (e1.e_addr & 7) + ((v1 + v2) << 3);
					if (v2 > 7) {
						++t2;
					}
					comma();
					t1 = addr(&e1);
					abscheck(&e2);
					if ((t2 == 0) && ((t1 == S_IDIX) || (t1 == S_IDIY))) {
					 	gixiy(t1);
						outab(0xCB);
						outrb(&e1,0);
						outab(op);
					 	break;
					}
				} else {
					op = (e1.e_addr & 7) + (v1 << 3);
					t1 = addr(&e1);
					if ((t1 == S_IDIX) || (t1 == S_IDIY)) {
				 		gixiy(t1);
						outab(0xCB);
						outrb(&e1,0);
						outab(op);
					 	break;
					 }
				}
				aerr();
				break;
			}
		}
		t2 = addr(&e2);
		if (t1 == S_R8) {
			if ((e1.e_addr == H) || (e1.e_addr == L)) {
				if ((t2 != S_R8IX) && (t2 != S_R8IY)) {
					v1 = op | e1.e_addr<<3;
					if (genop(0, v1, &e2, 0) == 0)break;
				}
			} else {
				v1 = op | e1.e_addr<<3;
				if (genop(0, v1, &e2, 0) == 0)break;
			}
			if (t2 == S_IMMED) {
				outab(e1.e_addr<<3 | 0x06);
				outrb(&e2,0);
				break;
			}
		}
		if ((t1 == S_R8IX) && (t2 != S_R8IY)) {
			if ((t2 != S_R8) || ((e2.e_addr != H) && (e2.e_addr != L))) {
				v1 = op | e1.e_addr<<3;
				if (genop(0xDD, v1, &e2, 0) == 0) break;
				if (t2 == S_IMMED) {
					outab(0xDD);
					outab(e1.e_addr<<3 | 0x06);
					outrb(&e2,0);
					break;
				}
			}
		}
		if ((t1 == S_R8IY) && (t2 != S_R8IX)) {
			if ((t2 != S_R8) || ((e2.e_addr != H) && (e2.e_addr != L))) {
				v1 = op | e1.e_addr<<3;
				if (genop(0xFD, v1, &e2, 0) == 0) break;
				if (t2 == S_IMMED) {
					outab(0xFD);
					outab(e1.e_addr<<3 | 0x06);
					outrb(&e2,0);
					break;
				}
			}
		}
		v1 = e1.e_addr;
		v2 = e2.e_addr;
		if ((t1 == S_R16) && (t2 == S_IMMED)) {
			v1 = gixiy(v1);
			outab(0x01|v1<<4);
			outrw(&e2, 0);
			break;
		}
if (target != TARGET_DMG) {
		if ((t1 == S_R16) && (t2 == S_INDM)) {
			if (gixiy(v1) == HL) {
				outab(0x2A);
			} else {
				outab(0xED);
				outab(0x4B | v1<<4);
			}
			outrw(&e2, 0);
			break;
		}
		if ((t1 == S_INDM) && (t2 == S_R16)) {
			if (gixiy(v2) == HL) {
				outab(0x22);
			} else {
				outab(0xED);
				outab(0x43 | v2<<4);
			}
			outrw(&e1, 0);
			break;
		}
		if ((t1 == S_R8) && (v1 == A) && (t2 == S_INDM)) {
			outab(0x3A);
			outrw(&e2, 0);
			break;
		}
		if ((t1 == S_INDM) && (t2 == S_R8) && (v2 == A)) {
			outab(0x32);
			outrw(&e1, 0);
			break;
		}
}
		if ((t2 == S_R8) && (gixiy(t1) == S_IDHL)) {
			outab(0x70|v2);
			if (t1 != S_IDHL)
				outrb(&e1, 0);
			break;
		}
		if ((t2 == S_IMMED) && (gixiy(t1) == S_IDHL)) {
			outab(0x36);
			if (t1 != S_IDHL)
				outrb(&e1, 0);
			outrb(&e2, 0);
			break;
		}
		if ((t1 == S_R8X) && (v1 != F) && (t2 == S_R8) && (v2 == A)) {
			outab(0xED);
			outab(v1);
			break;
		}
		if ((t1 == S_R8) && (v1 == A) && (t2 == S_R8X) && (v2 != F)) {
			outab(0xED);
			outab(v2|0x10);
			break;
		}
		if ((t1 == S_R16) && (v1 == SP)) {
			if ((t2 == S_R16) && (gixiy(v2) == HL)) {
				outab(0xF9);
				break;
			}
		}
		if ((t1 == S_R8) && (v1 == A)) {
			if ((t2 == S_IDBC) || (t2 == S_IDDE)) {
				outab(0x0A | (t2-S_INDR)<<4);
				break;
			}
		}
		if ((t2 == S_R8) && (v2 == A)) {
			if ((t1 == S_IDBC) || (t1 == S_IDDE)) {
				outab(0x02 | (t1-S_INDR)<<4);
				break;
			}
		}
if (target == TARGET_DMG) {
		/*
		 * 0x08 : LD (nn),SP
		 */
		if ((t1 == S_INDM) && (t2 == S_R16) && (v2 == SP)) {
			outab(0x08);
			outrw(&e1, 0);
			break;
		}
		/*
		 * 0xEA : LD (nn),A
		 * 0xFA : LD A,(nn)
		 */
		if ((t1 == S_INDM) && (t2 == S_R8) && (v2 == A)) {
			outab(0xEA);
			outrw(&e1, 0);
			break;
		}
		if ((t2 == S_INDM) && (t1 == S_R8) && (v1 == A)) {
			outab(0xFA);
			outrw(&e2, 0);
			break;
		}
		/*
		 * 0x32 : LD (HL-),A
		 * 0x3A : LD A,(HL-)
		 */
		if ((t1 == S_R8) && (v1 == A) && (t2 == S_IDHLD)) {
			outab(0x3A);
			break;
		}
		if ((t2 == S_R8) && (v2 == A) && (t1 == S_IDHLD)) {
			outab(0x32);
			break;
		}
		/*
		 * 0x22 : LD (HL+),A
		 * 0x2A : LD A,(HL+)
		 */
		if ((t1 == S_R8) && (v1 == A) && (t2 == S_IDHLI)) {
			outab(0x2A);
			break;
		}
		if ((t2 == S_R8) && (v2 == A) && (t1 == S_IDHLI)) {
			outab(0x22);
			break;
		}
}
		aerr();
		break;


	case S_EX:
		t1 = addr(&e1);
		comma();
		t2 = addr(&e2);
		if (t2 == S_R16) {
			v1 = e1.e_addr;
			v2 = e2.e_addr;
			if ((t1 == S_IDSP) && (v1 == 0)) {
				if (gixiy(v2) == HL) {
					outab(op);
					break;
				}
			}
			if (t1 == S_R16) {
				if ((v1 == DE) && (v2 == HL)) {
					outab(0xEB);
					break;
				}
			}
		}
		if ((t1 == S_R16X) && (t2 == S_R16X)) {
			outab(0x08);
			break;
		}
		aerr();
		break;

	case S_IN:
	case S_OUT:
		if (rf == S_IN) {
			t1 = addr(&e1);
			comma();
			t2 = addr(&e2);
			if ((t1 == S_R8X) && (e1.e_addr == F) && (t2 == S_IDC)) {
				outab(0xED);
				outab(0x40 + (6 << 3));
				break;
			}
		} else {
			t2 = addr(&e2);
			comma();
			t1 = addr(&e1);
			if (/*(target == TARGET_Z80) && */(t1 == S_IMMED) && (e1.e_addr == 0) && (t2 == S_IDC)) {
				outab(0xED);
				outab(0x41 + (6 << 3));
				break;
			}
			if (/*(target == TARGET_HD64) && */(t1 == S_IDHL) && (t2 == S_IDC)) {
				outab(0xED);
				outab(0x41 + (6 << 3));
				break;
			}
		}
		v1 = e1.e_addr;
		v2 = e2.e_addr;
		if (t1 == S_R8) {
			if ((v1 == A) && (t2 == S_INDM)) {
				outab(op);
				outab(v2);
				break;
			}
			if (t2 == S_IDC) {
				outab(0xED);
				outab(((rf == S_IN) ? 0x40 : 0x41) + (v1<<3));
				break;
			}
		}
		aerr();
		break;

	case S_DEC:
	case S_INC:
		t1 = addr(&e1);
		v1 = e1.e_addr;
		if (t1 == S_R8) {
			outab(op|(v1<<3));
			break;
		}
		if (t1 == S_R8IX) {
			outab(0xDD);
			outab(op|(v1<<3));
			break;
		}
		if (t1 == S_R8IY) {
			outab(0xFD);
			outab(op|(v1<<3));
			break;
		}
		if (t1 == S_IDHL) {
			outab(op|0x30);
			break;
		}
		if (t1 != gixiy(t1)) {
			outab(op|0x30);
			outrb(&e1,0);
			break;
		}
		if (t1 == S_R16) {
			v1 = gixiy(v1);
			if (rf == S_INC) {
				outab(0x03|(v1<<4));
				break;
			}
			if (rf == S_DEC) {
				outab(0x0B|(v1<<4));
				break;
			}
		}
		aerr();
		break;

	case S_JR:
		if ((v1 = admode(CND)) != 0) {
			if ((v1 &= 0xFF) <= 0x18) {
				op += (v1+1)<<3;
			} else {
				aerr();
			}
			comma();
		}
	case S_DJNZ:
		expr(&e2, 0);
		outab(op);
		if (mchpcr(&e2)) {
			v2 = e2.e_addr - dot.s_addr - 1;
			if ((v2 < -128) || (v2 > 127))
				aerr();
			outab(v2);
		} else {
			outrb(&e2, R_PCR);
		}
		if (e2.e_mode != S_USER)
			rerr();
		break;

	case S_CALL:
		if ((v1 = admode(CND)) != 0) {
			op |= (v1&0xFF)<<3;
			comma();
		} else {
			op = 0xCD;
		}
		expr(&e1, 0);
		outab(op);
		outrw(&e1, 0);
		break;

	case S_JP:
		if ((v1 = admode(CND)) != 0) {
			op |= (v1&0xFF)<<3;
			comma();
			expr(&e1, 0);
			outab(op);
			outrw(&e1, 0);
			break;
		}
		t1 = addr(&e1);
		if (t1 == S_USER) {
			outab(0xC3);
			outrw(&e1, 0);
			break;
		}
		if ((e1.e_addr == 0) && (gixiy(t1) == S_IDHL)) {
			outab(0xE9);
			break;
		}
		aerr();
		break;

	case X_Z80:
		target = TARGET_Z80;
		break;

	/*  R800 Instructions */
	case X_R800:
		target = TARGET_R800;
		break;
	case X_MULU:
		t1 = addr(&e1);
		if (more()) {
			comma();
			t2 = addr(&e2);
			if ((t1 == S_R8) && (e1.e_addr == A)) {
				if (t2 == S_R8) {
					outab(0xED);
					outab(op | (e2.e_addr << 3));
					break;
				}
			}
			if ((t1 == S_R16) && (e1.e_addr == HL)) {
				if (e2.e_addr <= SP) {
					outab(0xED);
					outab((2 + op) | (e2.e_addr << 4));
					break;
				}
			}
		} else {
			if ((t1 == S_R8) && (genop(0, op, &e1, 0)  == 0)) break;
		}
		aerr();
		break;

	/*  HD64180 Instructions */
	case X_HD64:
		target = TARGET_HD64;
		break;

	case X_INH2:
		outab(0xED);
		outab(op);
		break;

	case X_IN:
	case X_OUT:
		if (rf == X_IN) {
			t1 = addr(&e1);
			comma();
			t2 = addr(&e2);
			if ((t1 == S_R8X) && (e1.e_addr == F) && (t2 == S_INDM)) {
				outab(0xED);
				outab(0x30);
				outrb(&e2, 0);
				break;
			}
		} else {
			t2 = addr(&e2);
			comma();
			t1 = addr(&e1);
		}
		if ((t1 == S_R8) && (t2 == S_INDM)) {
			outab(0xED);
			outab(op | e1.e_addr<<3);
			outrb(&e2, 0);
			break;
		}
		aerr();
		break;

	case X_MLT:
		t1 = addr(&e1);
		if ((t1 == S_R16) && ((v1 = e1.e_addr) <= SP)) {
			outab(0xED);
			outab(op | v1<<4);
			break;
		}
		aerr();
		break;

	case X_TST:
		t1 = addr(&e1);
		if (t1 == S_R8) {
			outab(0xED);
			outab(op | e1.e_addr<<3);
			break;
		}
		if (t1 == S_IDHL) {
			outab(0xED);
			outab(0x34);
			break;
		}
		if (t1 == S_IMMED) {
			outab(0xED);
			outab(0x64);
			outrb(&e1, 0);
			break;
		}
		aerr();
		break;

	case X_TSTIO:
		t1 = addr(&e1);
		if (t1 == S_IMMED) {
			outab(0xED);
			outab(op);
			outrb(&e1, 0);
			break;
		}
		aerr();
		break;


	/* DMG-SYSTEM(GAMEBOY) Instructions */
	case X_DMG:
		target = TARGET_DMG;
		break;

	case X_STOP:	/* 0x10 */
		/*
		 * 0x10 : STOP
		 */
		outab(op);
		outab(0x00);
		break;

	case X_LDH:	/* 0xE0 */
		/*
		 * 0xE0 : LDH (n),A = LD ($FF00+n),A
		 * 0xE2 : LDH (C),A = LD ($FF00+C),A
		 * 0xF0 : LDH A,(n) = LD A,($FF00+n)
		 * 0xF2 : LDH A,(C) = LD A,($FF00+C)
		 */
		t1 = addr(&e1);
		comma();
		t2 = addr(&e2);
		if ((t1 == S_INDM) && (t2 == S_R8) && (e2.e_addr == A)) {
			outab(0xE0);
			outrb(&e1, 0);
			break;
		}
		if ((t1 == S_IDC) && (t2 == S_R8) && (e2.e_addr == A)) {
			outab(0xE2);
			break;
		}
		if ((t2 == S_INDM) && (t1 == S_R8) && (e1.e_addr == A)) {
			outab(0xF0);
			outrb(&e2, 0);
			break;
		}
		if ((t2 == S_IDC) && (t1 == S_R8) && (e1.e_addr == A)) {
			outab(0xF2);
			break;
		}
		aerr();
		break;

	case X_LDA:	/* 0xE8 */
		/*
		 * 0xE8 : LDA SP,#n(SP)
		 * 0xF8 : LDA HL,#n(SP)
		 */
		t1 = addr(&e1);
		comma();
		t2 = addr(&e2);
		if ((t1 == S_R16) && (e1.e_addr == SP) && (t2 == S_INDR+SP)) {
			outab(0xE8);
			outrb(&e2,0);
			break;
		}
		if ((t1 == S_R16) && (e1.e_addr == HL) && (t2 == S_INDR+SP)) {
			outab(0xF8);
			outrb(&e2,0);
			break;
		}
		aerr();
		break;

	case X_LDHL:	/* 0xF8 */
		/*
		 * 0xF8 : LDHL SP,#n
		 */
		t1 = addr(&e1);
		comma();
		t2 = addr(&e2);
		if ((t1 == S_R16) && (e1.e_addr == SP) && (t2 == S_IMMED)) {
			outab(op);
			outrb(&e2,0);
			break;
		}
		aerr();
		break;

	default:
		err('o');
	}
}

/*
 * general addressing evaluation
 * return(0) if general addressing mode output, else
 * return(esp->e_mode)
 */
int
genop(pop, op, esp, f)
register int pop, op;
register struct expr *esp;
int f;
{
	register int t1;
	if ((t1 = esp->e_mode) == S_R8) {
		if (pop) outab(pop);
		outab(op|esp->e_addr);
		return(0);
	}
	if ((t1 == S_R8IX) && (pop != 0xCB)) {
		outab(0xDD);
		if (pop && (pop != 0xDD)) outab(pop);
		outab(op|esp->e_addr);
		return(0);
	}
	if ((t1 == S_R8IY) && (pop != 0xCB)) {
		outab(0xFD);
		if (pop && (pop != 0xFD)) outab(pop);
		outab(op|esp->e_addr);
		return(0);
	}
	if (t1 == S_IDHL) {
		if (pop) outab(pop);
		outab(op|0x06);
		return(0);
	}
	if (gixiy(t1) == S_IDHL) {
		if (pop) {
			outab(pop);
			outrb(esp,0);
			outab(op|0x06);
		} else {
			outab(op|0x06);
			outrb(esp,0);
		}
		return(0);
	}
	if ((t1 == S_IMMED) && (f)) {
		if (pop) outab(pop);
		outab(op|0x46);
		outrb(esp,0);
		return(0);
	}
	return(t1);
}

/*
 * IX and IY prebyte check
 */
int
gixiy(v)
int v;
{
	if (target == TARGET_DMG) return(v);
	if (v == IX) {
		v = HL;
		outab(0xDD);
	} else if (v == IY) {
		v = HL;
		outab(0xFD);
	} else if (v == S_IDIX) {
		v = S_IDHL;
		outab(0xDD);
	} else if (v == S_IDIY) {
		v = S_IDHL;
		outab(0xFD);
	}
	return v;
}

/*
 * Branch/Jump PCR Mode Check
 */
int
mchpcr(esp)
register struct expr *esp;
{
	if (esp->e_base.e_ap == dot.s_area) {
		return(1);
	}
	if ((esp->e_flag==0) && (esp->e_base.e_ap==NULL)) {
		/*
		 * Absolute Destination
		 *
		 * Use the global symbol '.__.ABS.'
		 * of value zero and force the assembler
		 * to use this absolute constant as the
		 * base value for the relocation.
		 */
		esp->e_flag = 1;
		esp->e_base.e_sp = &sym[1];
	}
	return(0);
}

/*
 * The next character must be a
 * comma.
 */
int
comma()
{
	if (getnb() != ',')
		qerr();
	return(1);
}

/*
 * Machine dependent initialization
 */
VOID
minit()
{
	target = TARGET_Z80;
}
