#include <stdio.h>

enum {
	NOERROR = 0,
	ERROR_PARAM,
	ERROR_FILE_READ,
	ERROR_FILE_WRITE,
} BINCAT_ERRNO;

static void put_error(char *form, char *param) {
	fprintf(stderr, form, param);
}

static int bincat(FILE *ofp, char *path) {
	FILE *ifp;
	char buf[4096];
	int len, ret = NOERROR;
	ifp = fopen(path, "rb");
	if (ifp == NULL) {
		put_error("File %s read error\n", path);
		return ERROR_FILE_READ;
	}
	while (!feof(ifp)) {
		len = fread(buf, 1, sizeof(buf), ifp);
		if (ferror(ifp)) {
			put_error("File %s read error\n", path);
			ret = ERROR_FILE_READ;
			break;
		}
		if (len != fwrite(buf, 1, len, ofp)) {
			ret = ERROR_FILE_WRITE;
			break;
		}
	}
	fclose(ifp);
	return ret;
}

static int bincat_indirect(FILE *ofp, char *path)
{
	FILE *lfp;
	char fname[512];
	lfp = fopen(path, "r");
	if (lfp == NULL) {
		put_error("File %s read error\n", path);
		return ERROR_FILE_READ;
	}
	while (!feof(lfp) && fgets(fname, sizeof(fname), lfp)) {
		int l;
		l = strlen(fname);
		while (l > 0 && fname[l - 1] == '\n') fname[--l] = '\0';
		if (l > 0) {
			int ret;
			ret = bincat(ofp, fname);
			if (ret) return ret;
		}
	}
	return NOERROR;
}

int main(int argc, char **argv) {
	int i;
	FILE *ofp;
	if (argc < 3) {
		put_error("Usage: %s DEST SOURCE...\n", argv[0]);
		return ERROR_PARAM;
	}
	ofp = fopen(argv[1], "wb");
	if (ofp == NULL) {
		put_error("File %s write error\n", argv[1]);
		return ERROR_FILE_WRITE;
	}
	for (i = 2; i < argc; i++)
	{
		int ret;
		if (argv[i][0] == '@')
			ret = bincat_indirect(ofp, &argv[i][1]);
		else
			ret = bincat(ofp, argv[i]);
		if (ret) {
			if (ret == ERROR_FILE_WRITE)
				put_error("File %s write error\n", argv[1]);
			return ret;
		}
	}
	return NOERROR;
}
